;ATTiny24 with external 32 kHz (32768 Hz) low-frequency crystal
;Fuses: low = 0xC6, high = 0xDD, ext = 0xFF

;simple quartz clock for 4 7-segment LED displays (multiplexed)
;(C) Arne Rossius 2008-02-21, http://www.ebps.de.vu/

;PORTS:
;  PA0 = Anode for 10's of hours
;  PA1 = Anode for 1's of hours
;  PA2 = Anode for 10's of minutes
;  PA3 = Anode for 1's of minutes
;  PA4 = "Hours" key (to ground)
;  PA5 = "Minutes" key (to ground)
;  PA6 = Shift Register Data (Segment 'dp' first, then 'g' to 'a', active low)
;  PA7 = Shift Register Clock
;  PB0 = XTAL
;  PB1 = XTAL
;  PB2 = Shift Register Strobe

.equ	DOT_FREQ = 1 ;flashing dot: 0->1Hz, 1->.5Hz, 2->.25Hz, 7->static on
.equ	KEY_REPEATSTART = 16 ;16/16s = 1s until repeating starts
.equ	KEY_REPEAT = 5 ;16/5 = 3.2 repeats per second

.include "../include/tn24def.inc"
 
.def	disp1 = R0
.def	disp2 = R1
.def	disp3 = R2
.def	disp4 = R3
 
.def	temp = R16
.def	temp2 = R17
.def	temp3 = R18
.def	min1 = R19
.def	min10 = R20
.def	hr1 = R21
.def	hr10 = R22
.def	disp = R23
.def	sreg_data0 = R24
.def	sreg_data1 = R25
.def	repeat = R26

;R31:R30: Z pointer for Flash access

.equ	Hr = 4
.equ	Min = 5
.equ	Data = 6
.equ	Clock = 7
.equ	Strobe = 2

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org PCI0addr
	rjmp	keytoggle
.org OVF0addr
	rjmp	keyrepeat

;===============================================================================

disptab:
	.db 0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07
	.db 0x7f, 0x6f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

;===============================================================================

keytoggle:
	;pin change interrupt for hour/minute keys
	push	temp
	in	temp, SREG
	push	temp
	
	in	temp, PINA
	andi	temp, 0x30
	cpi	temp, 0x30
	breq	key_released
	
	;key pressed: reset and enable debounce/repeat timer
	ldi	temp, 0
	out	TCNT0, temp
	ldi	temp, 0x01 ;Clk/1 => ~8ms debouncing time
	out	TCCR0B, temp
	
keytoggle_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_released:
	;key released: disable debounce/repeat timer
	ldi	temp, 0x00
	out	TCCR0B, temp
	ldi	repeat, 0
	rjmp	keytoggle_end

;===============================================================================

keyrepeat:
	;key debounce/repeat timer overflow
	push	temp
	in	temp, SREG
	push	temp
	
	;change from debounce to repeat mode
	cpi	repeat, 0
	brne	repeatdiv
	ldi	temp, 0x02 ;Clk/8
	out	TCCR0B, temp
	ldi	temp, 0
	out	TCNT0, temp
	ldi	repeat, KEY_REPEATSTART
	rjmp	key_action
	
repeatdiv:
	;repeated keypress
	dec	repeat
	brne	keyrepeat_end
	ldi	repeat, KEY_REPEAT
key_action:
	;preform key action
	sbis	PINA, Hr
	rjmp	key_hr
	sbis	PINA, Min
	rjmp	key_min
	
keyrepeat_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_hr:
	;"Hour" key: increment hours
	rcall	inc_hr
	rcall	update_disp
	rjmp	keyrepeat_end

key_min:
	;"Minute" key: increment minutes and reset seconds
	rcall	inc_min
	rcall	update_disp
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	rjmp	keyrepeat_end
	
;===============================================================================

reset:
	;stack pointer setup
	ldi	temp, RAMEND
	out	SPL, temp
	
	;port setup
	ldi	temp, 0xCF
	out	DDRA, temp
	ldi	temp, 0x3F
	out	PORTA, temp
	ldi	temp, 0x04
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTB, temp
	
	;Timer 0 setup (Key debounce/repeat timer)
	ldi	temp, 1<<TOIE0
	out	TIMSK0, temp
	
	;Timer 1 setup (Clock timer)
	ldi	temp, 1<<WGM12 | 0x03 ;CTC Mode, Clk/64
	out	TCCR1B, temp
	ldi	temp, HIGH(30720) ;overflow every 60s (32768/(64*30720) = 1/60)
	out	OCR1AH, temp
	ldi	temp, LOW(30720)
	out	OCR1AL, temp
	
	;Pin Change interrupt setup
	ldi	temp, 0x30
	out	PCMSK0, temp
	ldi	temp, 1<<PCIE0
	out	GIMSK, temp
	
	;register setup
	ldi	hr10, 0
	ldi	hr1, 0
	ldi	min10, 0
	ldi	min1, 0
	rcall	update_disp
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	ldi	disp, 0x01
	
muxloop:
	;increment clock if timer reached 60s
	sbis	TIFR1, OCF1A
	rjmp	timeinc_end
	ldi	temp, 1<<OCF1A ;reset flag
	out	TIFR1, temp
	rcall	inc_min
	rcall	update_disp
timeinc_end:

	;get next display data
	sbrc	disp, 0
	mov	temp, disp1
	sbrc	disp, 1
	mov	temp, disp2
	sbrc	disp, 2
	mov	temp, disp3
	sbrc	disp, 3
	mov	temp, disp4
	
	;flashing dot
	ldi	temp2, 0x00
	in	temp3, TCNT1L ;need to read TCNT1L before TCNT1H can be read
	in	temp3, TCNT1H
	sbrs	temp3, DOT_FREQ
	ldi	temp2, 0x80
	sbrc	disp, 1
	or	temp, temp2
	
	;initialize registers for shifting
	in	sreg_data0, PORTA
	andi	sreg_data0, ~(1<<Data | 1<<Clock)
	in	sreg_data1, PORTA
	andi	sreg_data1, ~(1<<Data | 1<<Clock)
	ori	sreg_data1, 1<<Data
	
	;shift out 8 bits for next display
	out	PORTA, sreg_data0 ;Bit 7
	sbrs	temp, 7
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 6
	sbrs	temp, 6
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 5
	sbrs	temp, 5
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 4
	sbrs	temp, 4
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 3
	sbrs	temp, 3
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 2
	sbrs	temp, 2
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 1
	sbrs	temp, 1
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	out	PORTA, sreg_data0 ;Bit 0
	sbrs	temp, 0
	out	PORTA, sreg_data1
	sbi	PORTA, Clock
	
	;calculate next display anode selection mask
	com	disp
	andi	disp, 0x3F
	
	;switch to next display
	ldi	temp, 0x3F
	out	PORTA, temp
	sbi	PORTB, Strobe
	out	PORTA, disp
	cbi	PORTB, Strobe
	
	;restore value of 'disp'
	com	disp
	andi	temp, 0x0F
	
	lsl	disp
	brhs	PC+2
	rjmp	muxloop
	rjmp	loop
	
;===============================================================================

inc_min:
	;increment minutes
	inc	min1
	cpi	min1, 10
	brlo	inc_end
	clr	min1
	inc	min10
	cpi	min10, 6
	brlo	inc_end
	clr	min10
inc_hr:
	;increment hours
	inc	hr1
	cpi	hr1, 10
	brlo	hr_24
	clr	hr1
	inc	hr10
	rjmp	inc_end
hr_24:
	cpi	hr1, 4
	brne	inc_end
	cpi	hr10, 2
	brne	inc_end
	clr	hr1
	clr	hr10
inc_end:
	ret
	
;===============================================================================

update_disp:
	;convert time to 7-segment display representation
	ldi	ZH, 0
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr10
	lpm	disp1, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr1
	lpm	disp2, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, min10
	lpm	disp3, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, min1
	lpm	disp4, Z
	ret